"""
s3s3 version information.
"""
from subprocess import Popen, PIPE
from os.path import abspath, dirname


VERSION = (0, 1, 3, 'dev', 0)


def git_sha():
    loc = abspath(dirname(__file__))
    try:
        p = Popen(
            "cd \"%s\" && git log -1 --format=format:%%h" % loc,
            shell=True,
            stdout=PIPE,
            stderr=PIPE
        )
        return p.communicate()[0]
    except OSError:
        return None


def get_version(form='short'):
    """
    Returns the version string.
    Takes single argument ``form``, which should be one of the following
    strings:
    
    * ``short`` Returns major + minor branch version string with the format of
    B.b.t.
    * ``normal`` Returns human readable version string with the format of 
    B.b.t _type type_num.
    * ``verbose`` Returns a verbose version string with the format of
    B.b.t _type type_num@git_sha
    * ``all`` Returns a dict of all versions.
    """
    versions = {}
    branch = "%s.%s" % (VERSION[0], VERSION[1])
    tertiary = VERSION[2]
    type_ = VERSION[3]
    type_num = VERSION[4]
    
    versions["branch"] = branch
    v = versions["branch"]
    if tertiary:
        versions["tertiary"] = "." + str(tertiary)
        v += versions["tertiary"]
    versions['short'] = v
    if form is "short":
        return v
    v += " " + type_ + " " + str(type_num)
    versions["normal"] = v
    if form is "normal":
        return v
    v += " @" + git_sha()
    versions["verbose"] = v
    if form is "verbose":
        return v
    if form is "all":
        return versions



"""
The API for s3s3.
"""
import logging
import tempfile

from boto.s3.connection import S3Connection
from boto.s3.key import Key

from . import config


if config.pubsub['redis']:
    import redis
    r = redis.Redis()
else:
    Exception('Only redis is supported right now.')


def create_connection(connection_args):
    """
    Create a boto s3 connection using the connection_args dictionary.
    ``connection_args`` A dictionary of connection arguments.
    """
    connection_args = connection_args.copy()
    # bucket name is an invalid connection argument.
    connection_args.pop('bucket_name')
    connection_args.pop('verify_md5')
    return S3Connection(**connection_args)


def upload(source_key, dest_keys, verify_md5=False):
    """
    Download from the source s3 key to a named temporary file
    and upload to the destination s3 key.
    ``source_key`` The source boto s3 key.
    ``dest_keys`` A list of the destination boto s3 keys.
    ``verify_md5`` Verify md5 values when uploading. Source is always
    considered authoritative.
    """
    # Use the same name if no destination key is passed.
    if not dest_keys or not source_key:
        raise Exception(
            'The source_key and dest_keys parameters are required.')
    with tempfile.NamedTemporaryFile() as data:
        source_key.get_contents_to_file(data)
        data.file.flush()
        for dest_key in dest_keys:
            dest_key.set_contents_from_filename(data.name)
            _upload_verify_md5(verify_md5, source_key, dest_key)
            try:
                r.set(u'backup=>' + dest_key.key, True)
            except redis.ConnectionError:
                logging.warn('Unable to connect to redis')


def duplicate_bucket(source_bucket, dest_bucket, verify_md5=False):
    """
    Duplicate the source s3 bucket in the destination bucket.
    ``source_bucket`` The source boto s3 bucket.
    ``dest_bucket`` The destination boto s3 bucket.
    ``verify_md5`` Verify md5 values when duplicating. Source always
                   is considered authoritative. If md5 is missing this
                   will fetch the file to compute the md5 and make the
                   comparison.
    """
    for source_key in source_bucket.get_all_keys():
        dest_key = Key(dest_bucket)
        dest_key.key = source_key.key
        if not dest_key.exists():
            upload(source_key, [dest_key])
        if verify_md5:
            _update_md5([source_key, dest_key])
            if source_key.md5 != dest_key.md5:
                upload(source_key, [dest_key])
                logging.info('Uploaded {0} to {1} in bucket {2}.'.format(
                    dest_key.key,
                    dest_key.bucket.connection.host,
                    dest_key.bucket.name))


def _update_md5(keys):
    """
    If the md5 does not exist then retrieve the file so the md5 can be
    computed. This is required because some s3 clones do not provide
    md5 values.

    ``keys`` boto.s3.key.Key objects which may not have md5 hashes.
    """
    for key in keys:
        if not key.md5:
            with tempfile.NamedTemporaryFile() as data:
                key.get_contents_to_file(data)
                data.file.flush()
                logging.info('Updated md5 for {0} to'
                             ' {1} in bucket {2}.'.format(
                                 key.key,
                                 key.bucket.connection.host,
                                 key.bucket.name))


def _upload_verify_md5(verify_md5, source_key, dest_key):
    """
    If verify_md5 is True then log a message about the inconsistency.

    Don't raise an exception because it'll interrupt the blocking pubsub
    listen.
    """
    if verify_md5 and source_key.md5 != dest_key.md5:
        logging.warn('md5 hash does not match for {0} to'
                     ' {1} in bucket {2}.'.format(
                         dest_key.key,
                         dest_key.bucket.connection.host,
                         dest_key.bucket.name))



#!/usr/bin/env python
"""
Listens for s3 object names through the redis pubsub 'backup' channel.
When notified call the s3s3.api.upload function.

Configuration template: https://github.com/lsst-squre/s3s3/blob/master/s3s3/s3s3.ini.dist

Also: https://github.com/lsst-sqre/s3s3
"""
import argparse
import logging

from s3s3.client import ListenClient
from s3s3.config import initialize


def listen():
    """
    Subscribe to 'backup' redis pubsub channel and listen (a
    blocking call).
    """
    try:
        c = ListenClient()
        c.listen()
    except KeyboardInterrupt:
        return True
    except Exception as e:
        logging.warn(e)
        return False


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--config', type=str,
                        help='Configuration file to use.')
    args = parser.parse_args()
    if args.config:
        initialize(args.config)
    if listen():
        exit(0)
    else:
        exit(1)


if __name__ == "__main__":
    main()



from .. import config
from .redis import Listener

def get_listen():
    if config.pubsub['redis']:
        listener = Listener('backup')
        return listener.run
    if config.pubsub['sns']:
        return None



"""
Clients that use the config to provide access to the api.
"""
from functools import partial
import logging

from boto.s3.key import Key

from .api import create_connection, upload, duplicate_bucket
from . import config
from .pubsub import get_listen


def on_notify(source_conn, dest_conns, s3_key):
    """
    Replicate the s3_key from a source s3 connection to s3 destinations.
    """
    # s3_key is of type bytes and boto requires a str.
    s3_key = s3_key.decode('utf-8')
    source_key = _get_key(source_conn.bucket, s3_key)
    if source_key.exists():
        dest_keys = []
        dest_keys_verify_md5 = False
        for dest_name, dest_conn in dest_conns.items():
            dest_key = _get_key(dest_conn.s3s3_bucket, s3_key)
            if not dest_key.exists():
                dest_keys.append(dest_key)
                dest_keys_verify_md5 |= dest_conn.s3s3_verify_md5
        upload(source_key,
               dest_keys,
               self.source_conn.s3s3_verify_md5 or dest_keys_verify_md5)
    else:
        logging.warn('s3_key: {0} does not exist in '
                     'source s3 bucket.'.format(s3_key))


class Client(object):
    """
    A base client that configures the source and destination connections.
    """

    def __init__(self):
        self.source_conn = create_connection(config.source)
        self.source_conn.s3s3_verify_md5 = config.source['verify_md5']
        self.source_conn.s3s3_bucket = _get_bucket(self.source_conn)
        self.dest_conns = {name: create_connection(d)\
                           for name, d in config.destinations.items()}
        for name, dc in self.dest_conns.items():
            dc.s3s3_bucket = _get_bucket(dc, name)
            dc.s3s3_verify_md5 = config.destinations[name]['verify_md5']


class ListenClient(Client):
    """
    A client to listen for 'backup' redis pubsub channels and call
    on_notify.
    """

    def listen(self):
        _listen = get_listen()
        _listen(partial(on_notify, self.source_conn, self.dest_conns))


class BucketClient(Client):
    """
    A client that uses the configuration to provide access
    to the duplicate_bucket API function.

    Duplicate all s3 objects in the config.source s3 bucket to
    all destination buckets.
    """

    def duplicate(self):
        """
        Duplicate the source s3 bucket in the destination s3 bucket.
        """
        try:
            source_bucket = self.source_conn.s3s3_bucket
            source_verify_md5 = self.source_conn.s3s3_verify_md5
            for _, dest_conn in self.dest_conns.items():
                dest_verify_md5 = dest_conn.s3s3_verify_md5
                duplicate_bucket(
                    source_bucket,
                    dest_conn.s3s3_bucket,
                    verify_md5=(source_verify_md5 or dest_verify_md5))
        except Exception as e:
            logging.error(e)
            return False
        return True


def _get_bucket(conn, name=None):
    """
    Return the bucket for the connection.

    If there is a name then it's a destination connection. Otherwise
    it is the source connection.
    """
    if name:
        return conn.get_bucket(config.destinations[name]['bucket_name'])
    else:
        return conn.get_bucket(config.source['bucket_name'])


def _get_key(bucket, s3_key):
    """
    Create a boto.s3.key.Key using the bucket and s3_key (name).
    """
    key = Key(bucket)
    key.key = s3_key
    return key



import configparser
from distutils.util import strtobool
import importlib
import logging
import os


required_sections = ['source',
                     'destinations',
                     'pubsub']
source = None
destinations = None
pubsub = None
config_files = ['/usr/local/etc/s3s3.ini',
                os.path.join(os.path.dirname(__file__), 's3s3.ini'),
                os.path.join(os.pardir, 's3s3.ini'),
                os.path.join(os.curdir, 's3s3.ini')]


def get_source_connection(config):
    if config.has_section('source'):
        return {k: v for k, v in config['source'].items()}
    else:
        return {}

def get_dest_connections(config):
    destinations = {}
    for section_name in config.sections():
        if section_name.startswith('dest'):
            destinations[section_name] =\
                {k: v for k, v in config[section_name].items()}
    return destinations


def get_calling_format(name):
    """
    Get the calling format class from the boto.s3.connection module
    and return an instance of it.
    """
    botos3connection = importlib.import_module('boto.s3.connection')
    calling_format_class = getattr(botos3connection, name)
    return calling_format_class()


def connection_fixup(source, destinations):
    """
    Fix the configuration connection dictionaries so they
    can be passed in to boto to create valid boto s3 connections.
    """
    for conn_dicts in [source, destinations]:
        for _, conn in conn_dicts.items():
            if conn.get('verify_md5'):
                conn['verify_md5'] = bool(strtobool(conn['verify_md5']))
            else:
                conn['verify_md5'] = False
            if conn.get('is_secure'):
                conn['is_secure'] = bool(strtobool(conn['is_secure']))
            if conn.get('calling_format'):
                conn['calling_format'] =\
                get_calling_format(conn['calling_format'])


def get_pubsub(config):
    if config.has_section('pubsub'):
        pubsub = {k: v for k, v in config['pubsub'].items()}
        pubsub['redis'] = bool(strtobool(pubsub.get('redis')))
        pubsub['sns'] = bool(strtobool(pubsub.get('sns')))
        if (pubsub.get('redis') ^ pubsub.get('sns')):
            return pubsub
        else:
            logging.warn('Use either redis or sns for pubsub but not both.')
            raise Exception('Use either redis or sns for pubsub but not both.')
    else:
        return {'redis': False, 'sns': False}


def has_required_sections(config):
    sections = config.sections()
    for section in required_sections:
        if not (section in sections):
            return False
    return True


def initialize(config_file=None):
    global source, destinations, pubsub
    config = configparser.ConfigParser()
    if config_file:
        config_files.append(os.path.expanduser(config_file))
    files_read = config.read(config_files)
    if has_required_sections(config):
        logging.warn('Missing required sections: {0}'.format(
            required_sections))
    source = get_source_connection(config)
    destinations = get_dest_connections(config)
    pubsub = get_pubsub(config)
    connection_fixup({'source': source}, destinations)


initialize()



"""
Redis pubsub.
"""
import logging
import threading

import redis


class Listener(threading.Thread):
    """
    A class to encapsulate redis pubsub.
    """
    def __init__(self, channels):
        threading.Thread.__init__(self)
        self.redis = redis.Redis()
        self.pubsub = self.redis.pubsub()
        self.pubsub.subscribe(channels)

    def run(self, method):
        """
        A blocking function.
        ``method`` A method that takes the message data field as a parameter.
        """
        try:
            for message in self.pubsub.listen():
                if message['type'] == 'message':
                    method(message['data'])
        except redis.ConnectionError as ce:
            logging.warn(ce)



import os
import setuptools
from s3s3.version import get_version

readme = open('README.md').read()
requirements = open('requirements.txt').read().split('\n')

long_description = """
s3s3 %s
A microservice to move files from S3 APIs (Swift or Ceph) to other S3 APIs.

To install use `pip install s3s3`

----

%s

----

For more information, please see: https://github.com/lsst-sqre/s3s3
""" % (get_version('short'), readme)

setuptools.setup(
    name='s3s3',
    version=get_version('short'),
    author='jmatt',
    author_email='jmatt@lsst.org',
    description="A microservice to move files from S3 APIs (Swift or Ceph) to other S3 APIs.",
    long_description=long_description,
    url="https://github.com/lsst-sqre/s3s3",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries",
        "Topic :: System",
        "Topic :: System :: Archiving :: Backup",
        "Topic :: System :: Systems Administration"
    ],
    entry_points={
        'console_scripts': [
            's3s3listen = s3s3.scripts.listen:main',
            's3s3bucket = s3s3.scripts.bucket:main'
        ]
    })
